async function sha256(text){
  const uint8  = new TextEncoder().encode(text)
  const digest = await crypto.subtle.digest('SHA-256', uint8)
  return Array.from(new Uint8Array(digest)).map(v => v.toString(16).padStart(2,'0')).join('')
}

async function autoRun(){
  let run_run_option = 0;

  let run2_run_option = 0;
  let run2_latency1 = null;
  let run2_latency2 = null;
  
  let run3_run_option = 0;
  let run3_wpm = null;
  
  let run4_run_option = 0;
  let run4_wpm1 = null;
  let run4_wpm2 = null;
  
  let run5_run_option = 0;
  
  let run6_run_option = 0;
  let run6_target = 1;
  
  let run7_run_option = 0;
  let run7_target = 1;

  let items = await chrome.storage.local.get([
      'run_run_option',
      'run2_run_option',
      'run2_latency1',
      'run2_latency2',
      'run3_run_option',
      'run3_wpm',
      'run4_run_option',
      'run4_wpm1',
      'run4_wpm2',
      'run5_run_option',
      'run6_run_option',
      'run6_target',
      'run7_run_option',
      'run7_target',
    ]);

  if(items.run_run_option){
    run_run_option = items.run_run_option;
  }
  if(items.run2_run_option){
    run2_run_option = items.run2_run_option;
  }
  if(items.run2_latency1){
    run2_latency1 = Number(items.run2_latency1);
  }
  if(items.run2_latency2){
    run2_latency2 = Number(items.run2_latency2);
  }
  if(items.run3_run_option){
    run3_run_option = items.run3_run_option;
  }
  if(items.run3_wpm){
    run3_wpm = Number(items.run3_wpm);
  }
  if(items.run4_run_option){
    run4_run_option = items.run4_run_option;
  }
  if(items.run4_wpm1){
    run4_wpm1 = Number(items.run4_wpm1);
  }
  if(items.run4_wpm2){
    run4_wpm2 = Number(items.run4_wpm2);
  }
  if(items.run5_run_option){
    run5_run_option = items.run5_run_option;
  }
  if(items.run6_run_option){
    run6_run_option = items.run6_run_option;
  }
  if(items.run6_target){
    run6_target = items.run6_target;
  }
  if(items.run7_run_option){
    run7_run_option = items.run7_run_option;
  }
  if(items.run7_target){
    run7_target = items.run7_target;
  }
  
  const dataList = getInfo();
  let dataList2 = [];

  if(run_run_option === '1'){
    console.log('--- run auto run is detected ---');
    writeResult(dataList);
  }

  if(run2_run_option === '1'){
    console.log('--- run2 auto run is detected ---');
    dataList2 = changeLatency(dataList, run2_latency1, run2_latency2);
    writePrev(dataList2);
  }
  
  if(run3_run_option === '1'){
    console.log('--- run3 auto run is detected ---');
    dataList2 = cutHighestWPM(dataList, run3_wpm);
    writePrev(dataList2);
  }

  if(run4_run_option === '1'){
    console.log('--- run4 auto run is detected ---');
    dataList2 = cutLowestWPM(dataList, run4_wpm1, run4_wpm2);
    writePrev(dataList2);
  }

  if(run5_run_option === '1'){
    console.log('--- run5 auto run is detected ---');
    recordResult(dataList);
  }

  if(run6_run_option === '1'){
    console.log('--- run6 auto run is detected ---');
    writeResult2(dataList, run6_target);
  }

  if(run7_run_option === '1'){
    console.log('--- run7 auto run is detected ---');
    changeWPM(dataList, run7_target).then(dataList2 => {writePrev(dataList2);});
  }

}

function getInfo(){
  // DOMからe-typingの各ワードのリザルトを取得
  const info = document.getElementById('typing_content').contentDocument.getElementById('exampleList').children[0];
  let data_list = [];
  for (let i = 0; i < 15; i++){
    if (info.children[i].children.length <= 2 || info.children[i].children[2].innerHTML[0] !== 'l'){
      break;
    }
    const word1 = info.children[i].children[0].innerHTML;
    const word2_all = info.children[i].children[1].children;
    let word2 = '';
    for (let j = 0; j < word2_all.length; j++){
      if (word2_all[j].getAttribute('data-tooltip')){
        word2 += word2_all[j].innerHTML;
      }
    }
    let is_word2_finished = true;
    if(word2.length != word2_all.length){
      is_word2_finished = false;
    }

    const stats = info.children[i].children[2].innerHTML;
    const latency = Number(stats.split(',')[0].split(': ')[1]);
    const rkpm = Number(stats.split(',')[2].split(': ')[1]);
    const miss = Number(stats.split(',')[3].split(': ')[1]);
    const data = [word1, word2, latency, rkpm, miss, is_word2_finished, word2_all.length];
    data_list.push(data);
  }

  return data_list;
}

function writeResult(dataList){
  const info = document.getElementById('typing_content').contentDocument.getElementById('exampleList').children[0];
  let total_length = 0;
  let total_time = 0;
  let total_miss = 0;
  const data_list = dataList;

  // 取得した情報を処理、リザルト画面に書き込み
  for (let i = 0; i < data_list.length; i++){
    let word1 = data_list[i][0];
    let word2 = data_list[i][1];
    let latency = data_list[i][2];
    let rkpm = data_list[i][3];
    let miss = data_list[i][4];

    total_length += word2.length;
    total_time += latency + 60 * (word2.length - 1) / rkpm;
    total_miss += miss;

    var elem = document.createElement('div');
    var wpm = 60 * (total_length / total_time);
    var accuracy = (total_length - total_miss) / total_length;
    var score = wpm * accuracy * accuracy * accuracy;
    elem.innerHTML = (i + 1) + '番目までのWPM: ' + (Math.floor(wpm * 100) / 100) + '<br>' + (i + 1) + '番目までのスコア: ' + (Math.floor(score * 100) / 100);
    info.children[i].append(elem);
  }

  return true;
}

async function writeResult2(dataList, target){
  const info = document.getElementById('typing_content').contentDocument.getElementById('exampleList').children[0];
  const data_list = dataList;

  for (let i = 0; i < data_list.length; i++){
    let word1 = data_list[i][0];
    let word1_sha256 = await sha256(word1);
    let word2 = data_list[i][1];
    let is_word2_finished = data_list[i][5];
    let latency = data_list[i][2];
    let rkpm = data_list[i][3];
    let miss = data_list[i][4];

    let time = latency + 60 * (word2.length - 1) / rkpm;
    let single_wpm = 60 * (word2.length / time);

    let records_dic = await chrome.storage.local.get([word1_sha256]);
    let word_records = records_dic[word1_sha256];
    if(word_records == null){
      word_records = [];
    }
    if(is_word2_finished){
      word_records.push([word1, word2, latency, rkpm, miss, Date.now()])
    }
    if(word_records.length == 0){
      continue;
    }

    let word_records_new = word_records.map(record => {
      let word2 = record[1];
      let latency = record[2];
      let rkpm = record[3];
      let time = latency + 60 * (word2.length - 1) / rkpm;
      let single_wpm = 60 * (word2.length / time);
      return [...record, single_wpm];
    })

    //console.log(word_records_new);

    if(target === '1'){
      word_records_new.sort((a,b) => {return (b[3] - a[3]);});
      let word_records_size = word_records_new.length;
      let wpm_max = word_records_new[0][3];
      let wpm_median = 0;
      if(word_records_size % 2 == 0){
        wpm_median += word_records_new[word_records_size / 2 - 1][3];
        wpm_median += word_records_new[word_records_size / 2][3];
        wpm_median /= 2;
      }
      else{
        wpm_median = word_records_new[(word_records_size - 1) / 2][3];
      }
      let wpm_rank = '-';
      for(let i = 0; i < word_records_size; i++){
        if(word_records_new[i][3] == rkpm){
          wpm_rank = i + 1;
          break;
        }
      }
  
      let elem = document.createElement('div');
      elem.innerHTML = 'rkpm(best): ' + wpm_max.toFixed(0) + ', rkpm(median): ' + wpm_median.toFixed(0) + ', rank: ' + wpm_rank.toString() + '/' + word_records_size.toString();
      elem.style.fontSize = '12px';
      elem.style.position = 'relative';
      elem.style.top = '-2px';
      info.children[i].children[2].after(elem);
    }
    else if(target === '2'){
      word_records_new.sort((a,b) => {return (b[6] - a[6]);});
      let word_records_size = word_records_new.length;
      let wpm_max = word_records_new[0][6];
      let wpm_median = 0;
      if(word_records_size % 2 == 0){
        wpm_median += word_records_new[word_records_size / 2 - 1][6];
        wpm_median += word_records_new[word_records_size / 2][6];
        wpm_median /= 2;
      }
      else{
        wpm_median = word_records_new[(word_records_size - 1) / 2][6];
      }
      let wpm_rank = '-';
      for(let i = 0; i < word_records_size; i++){
        if(word_records_new[i][6] == single_wpm){
          wpm_rank = i + 1;
          break;
        }
      }
  
      let elem = document.createElement('div');
      elem.innerHTML = 'WPM(best): ' + wpm_max.toFixed(0) + ', WPM(median): ' + wpm_median.toFixed(0) + ', rank: ' + wpm_rank.toString() + '/' + word_records_size.toString();
      elem.style.fontSize = '12px';
      elem.style.position = 'relative';
      elem.style.top = '-2px';
      info.children[i].children[2].after(elem);
    }
  }
}

function changeLatency(dataList, latency1, latency2){
  const data_list = dataList;
  let score = 0;
  let WPM = 0;
  let latency_all = 0;

  let total_length = 0;
  let total_time = 0;
  let total_miss = 0;

  if(latency2 == 0){
    for (let i = 0; i < data_list.length; i++){
      let word1 = data_list[i][0];
      let word2 = data_list[i][1];
      let latency = data_list[i][2];
      if(word2[0] == 'A' || word2[0] == 'S' || word2[0] == 'D' || word2[0] == 'F' || word2[0] == 'J' || word2[0] == 'K'){
        latency = latency1;
      }
      else{
        latency = latency1;
      }
      let rkpm = data_list[i][3];
      let miss = data_list[i][4];

      total_length += word2.length;
      total_time += latency + 60 * (word2.length - 1) / rkpm;
      total_miss += miss;
      latency_all += latency;
    }
  }
  else{
    for (let i = 0; i < data_list.length; i++){
      let word1 = data_list[i][0];
      let word2 = data_list[i][1];
      let latency = data_list[i][2];
      if(word2[0] == 'A' || word2[0] == 'S' || word2[0] == 'D' || word2[0] == 'F' || word2[0] == 'J' || word2[0] == 'K'){
        latency = latency2;
      }
      else{
        latency = latency1;
      }
      let rkpm = data_list[i][3];
      let miss = data_list[i][4];

      total_length += word2.length;
      total_time += latency + 60 * (word2.length - 1) / rkpm;
      total_miss += miss;
      latency_all += latency;
    }
  }
  let accuracy = (total_length - total_miss) / total_length;
  WPM = 60 * (total_length / total_time);
  score = WPM * accuracy * accuracy * accuracy;
  latency_all = latency_all / data_list.length;
  WPM = WPM.toFixed(2)
  score = Math.floor(score);
  latency_all = latency_all.toFixed(3);

  return [score, WPM, latency_all];
}

function cutHighestWPM(dataList, n){
  const data_list = dataList;
  let score = 0;
  let WPM = 0;
  let latency_all = 0;

  let total_length = 0;
  let total_time = 0;
  let total_miss = 0;

  let n_number = n;
  let cutoff_number = Math.min(data_list.length, n_number);
  let data_list_new = [];

  for (let i = 0; i < data_list.length; i++){
    let word1 = data_list[i][0];
    let word2 = data_list[i][1];
    let latency = data_list[i][2];
    let rkpm = data_list[i][3];
    let miss = data_list[i][4];

    total_length += word2.length;
    total_miss += miss;
    latency_all += latency;

    let time = latency + 60 * (word2.length - 1) / rkpm;
    let single_wpm = 60 * (word2.length / time);
    data_list_new.push([word1, word2, latency, rkpm, miss, single_wpm]);
  }

  data_list_new.sort((a,b) => {return (b[5] - a[5]);});
  data_list_new.slice(0,cutoff_number);

  let total_length_2 = 0;
  let total_time_2 = 0;

  for (let i = 0; i < cutoff_number; i++){
    let word1 = data_list_new[i][0];
    let word2 = data_list_new[i][1];
    let latency = data_list_new[i][2];
    let rkpm = data_list_new[i][3];
    let miss = data_list_new[i][4];

    total_length_2 += word2.length;
    total_time_2 += latency + 60 * (word2.length - 1) / rkpm;
  }
  let accuracy = (total_length - total_miss) / total_length;
  WPM = 60 * (total_length_2 / total_time_2);
  score = WPM * accuracy * accuracy * accuracy;
  latency_all = latency_all / data_list.length;
  WPM = WPM.toFixed(2)
  score = Math.floor(score);
  latency_all = latency_all.toFixed(3);

  return [score, WPM, latency_all];
}

function cutLowestWPM(dataList, wpm1, wpm2){
  const data_list = dataList;
  let score = 0;
  let WPM = 0;
  let latency_all = 0;

  let total_length = 0;
  let total_time = 0;
  let total_miss = 0;

  if(wpm2 == 0){
    for (let i = 0; i < data_list.length; i++){
      let word1 = data_list[i][0];
      let word2 = data_list[i][1];
      let latency = data_list[i][2];
      let rkpm = data_list[i][3];
      let miss = data_list[i][4];
  
      total_length += word2.length;
      total_miss += miss;
      latency_all += latency;
  
      let time = latency + 60 * (word2.length - 1) / rkpm;
      let single_wpm = 60 * (word2.length / time);
      if(single_wpm < wpm1){
        total_time += (60 * word2.length) / wpm1;
      }
      else{
        total_time += time;
      }
    }
  }
  else{
    for (let i = 0; i < data_list.length; i++){
      let word1 = data_list[i][0];
      let word2 = data_list[i][1];
      let latency = data_list[i][2];
      let rkpm = data_list[i][3];
      let miss = data_list[i][4];

      total_length += word2.length;
      total_miss += miss;
      latency_all += latency;

      let time = latency + 60 * (word2.length - 1) / rkpm;
      let single_wpm = 60 * (word2.length / time);
      if(single_wpm < wpm1){
        total_time += (60 * word2.length) / wpm2;
      }
      else{
        total_time += time;
      }
    }
  }
  let accuracy = (total_length - total_miss) / total_length;
  WPM = 60 * (total_length / total_time);
  score = WPM * accuracy * accuracy * accuracy;
  latency_all = latency_all / data_list.length;
  WPM = WPM.toFixed(2)
  score = Math.floor(score);
  latency_all = latency_all.toFixed(3);

  return [score, WPM, latency_all];
}

async function changeWPM(dataList, target){
  const data_list = dataList;
  let score = 0;
  let WPM = 0;
  let latency_all = 0;

  let total_length = 0;
  let total_time = 0;
  let total_miss = 0;

  for (let i = 0; i < data_list.length; i++){
    let word1 = data_list[i][0];
    let word1_sha256 = await sha256(word1);
    let word2 = data_list[i][1];
    let is_word2_finished = data_list[i][5];
    let latency = data_list[i][2];
    let rkpm = data_list[i][3];
    let miss = data_list[i][4];
    let word2_all = data_list[i][6];

    let records_dic = await chrome.storage.local.get([word1_sha256]);
    let word_records = records_dic[word1_sha256];
    if(word_records == null){
      word_records = [];
    }
    if(is_word2_finished){
      word_records.push([word1, word2, latency, rkpm, miss, Date.now()])
    }
    if(word_records.length == 0){
      break;
    }

    total_length += word2_all;
    total_miss += miss;
    latency_all += latency;

    let word_records_new = word_records.map(record => {
      let word2 = record[1];
      let latency = record[2];
      let rkpm = record[3];
      let time = latency + 60 * (word2.length - 1) / rkpm;
      let single_wpm = 60 * (word2.length / time);
      return [...record, single_wpm];
    })

    word_records_new.sort((a,b) => {return (b[6] - a[6]);});
    let word_records_size = word_records_new.length;
    let wpm_max = word_records_new[0][6];
    let wpm_median = 0;
    if(word_records_size % 2 == 0){
      wpm_median += word_records_new[word_records_size / 2 - 1][6];
      wpm_median += word_records_new[word_records_size / 2][6];
      wpm_median /= 2;
    }
    else{
      wpm_median = word_records_new[(word_records_size - 1) / 2][6];
    }
    
    let time = 0;
    if(target === '1'){
      time = (60 * word2_all) / wpm_median;
    }
    else if(target === '2'){
      time = (60 * word2_all) / wpm_max;
    }
    total_time += time;
  }

  let accuracy = (total_length - total_miss) / total_length;
  WPM = 60 * (total_length / total_time);
  score = WPM * accuracy * accuracy * accuracy;
  latency_all = latency_all / data_list.length;
  WPM = WPM.toFixed(2)
  score = Math.floor(score);
  latency_all = latency_all.toFixed(3);

  return [score, WPM, latency_all];
}

function judgeLevel(data){
  if(data <= 21){
    return 'E-';
  }
  else if(data <= 38){
    return 'E';
  }
  else if(data <= 55){
    return 'E+';
  }
  else if(data <= 72){
    return 'D-';
  }
  else if(data <= 89){
    return 'D';
  }
  else if(data <= 106){
    return 'D+';
  }
  else if(data <= 123){
    return 'C-';
  }
  else if(data <= 140){
    return 'C';
  }
  else if(data <= 157){
    return 'C+';
  }
  else if(data <= 174){
    return 'B-';
  }
  else if(data <= 191){
    return 'B';
  }
  else if(data <= 208){
    return 'B+';
  }
  else if(data <= 225){
    return 'A-';
  }
  else if(data <= 242){
    return 'A';
  }
  else if(data <= 259){
    return 'A+';
  }
  else if(data <= 276){
    return 'S';
  }
  else if(data <= 299){
    return 'Good!';
  }
  else if(data <= 324){
    return 'Fast';
  }
  else if(data <= 349){
    return 'Thunder';
  }
  else if(data <= 374){
    return 'Ninja';
  }
  else if(data <= 399){
    return 'Comet';
  }
  else if(data <= 449){
    return 'Professor';
  }
  else if(data <= 499){
    return 'LaserBeam';
  }
  else if(data <= 549){
    return 'EddieVH';
  }
  else if(data <= 599){
    return 'Meijin';
  }
  else if(data <= 649){
    return 'Rocket';
  }
  else if(data <= 699){
    return 'Tatujin';
  }
  else if(data <= 749){
    return 'Jedi';
  }
  else if(data <= 799){
    return 'Godhand';
  }
  else if(data <= 1100){
    return 'Joker';
  }
  else{
    return 'Error';
  }
}

function judgeLevel2(data){
  if(data <= 100){
    return '一般初心者';
  }
  else if(data <= 150){
    return '一般初級者';
  }
  else if(data <= 200){
    return '一般中級者';
  }
  else if(data <= 250){
    return '一般上級者';
  }
  else if(data <= 300){
    return '一般最上級者';
  }
  else if(data <= 350){
    return 'タイパー初心者';
  }
  else if(data <= 400){
    return 'タイパー初級者';
  }
  else if(data <= 450){
    return 'タイパー中級者';
  }
  else if(data <= 500){
    return 'タイパー上級者';
  }
  else if(data <= 550){
    return 'タイパー最上級者';
  }
  else if(data <= 600){
    return '変態パー初心者';
  }
  else if(data <= 650){
    return '変態パー初級者';
  }
  else if(data <= 700){
    return '変態パー中級者';
  }
  else if(data <= 750){
    return '変態パー上級者';
  }
  else if(data <= 800){
    return '変態パー最上級者';
  }
  else if(data <= 850){
    return '変態パー絶対神';
  }
  else if(data <= 900){
    return '変態パー究極神';
  }
  else if(data <= 950){
    return '変態パー超越神';
  }
  else if(data > 950){
    return '変態パー破壊神';
  }
  else{
    return '-';
  }
}

function writePrev(dataList){
  const info = document.getElementById('typing_content').contentDocument.getElementById('prev').children[1];
  let score = dataList[0];
  let WPM = dataList[1];
  let latency = dataList[2];

  info.children[0].children[0].children[0].innerHTML = score; // score
  info.children[0].children[1].children[0].innerHTML = judgeLevel(score); // Level
  info.children[0].children[2].children[0].innerHTML = ''; // total time
  info.children[0].children[3].children[0].innerHTML = ''; // total char
  info.children[0].children[4].children[0].innerHTML = ''; // miss char
  info.children[0].children[5].children[0].innerHTML = WPM; // WPM
  info.children[0].children[6].children[0].innerHTML = ''; // Accuracy
  info.children[0].children[7].children[0].innerHTML = latency; // latency
  info.children[0].children[8].children[0].innerHTML = ''; // rkpm

  return true;
}

async function recordResult(dataList){
  const data_list = dataList;

  for (let i = 0; i < data_list.length; i++){
    let word1 = data_list[i][0];
    let word1_sha256 = await sha256(word1);
    let word2 = data_list[i][1];
    let is_word2_finished = data_list[i][5];
    let latency = data_list[i][2];
    let rkpm = data_list[i][3];
    let miss = data_list[i][4];

    if(is_word2_finished){
      let records_dic = await chrome.storage.local.get([word1_sha256]);
      let word_records = records_dic[word1_sha256];
      if(word_records == null){
        word_records = [];
      }
      word_records.push([word1,word2,latency,rkpm,miss,Date.now()]);
      chrome.storage.local.set({
        [word1_sha256]: word_records,
      });
    }
  }

  return true;

}

chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
  let dataList = [];
  dataList = getInfo();

  let dataList2 = [];
  //console.log(dataList);

  switch(request.msg) {
    case 'run':
      console.log('--- run is activated ---\n');
      const run_result = writeResult(dataList);
      if(run_result){
        sendResponse(run_result);
      }
      else{
        sendResponse('--- Error is detected ---\n');
      }
      break;
    case 'run2':
      console.log('--- run2 is activated ---\n');
      dataList2 = changeLatency(dataList, Number(request.latency1), Number(request.latency2));
      const run2_result = writePrev(dataList2);
      if(run2_result){
        sendResponse(run2_result);
      }
      else{
        sendResponse('--- Error is detected ---\n');
      }
      break;
    case 'run3':
      console.log('--- run3 is activated ---\n');
      dataList2 = cutHighestWPM(dataList, Number(request.wpm));
      const run3_result = writePrev(dataList2);
      if(run3_result){
        sendResponse(run3_result);
      }
      else{
        sendResponse('--- Error is detected ---\n');
      }
      break;
    case 'run4':
      console.log('--- run4 is activated ---\n');
      dataList2 = cutLowestWPM(dataList, Number(request.wpm1), Number(request.wpm2));
      const run4_result = writePrev(dataList2);
      if(run4_result){
        sendResponse(run4_result);
      }
      else{
        sendResponse('--- Error is detected ---\n');
      }
      break;
    case 'run5':
      console.log('--- run5 is activated ---\n');
      const run5_result = recordResult(dataList);
      if(run5_result){
        sendResponse(run5_result);
      }
      else{
        sendResponse('--- Error is detected ---\n');
      }
      break;
    case 'run6':
      console.log('--- run6 is activated ---\n');
      const run6_result = writeResult2(dataList, request.target);
      if(run6_result){
        sendResponse(run6_result);
      }
      else{
        sendResponse('--- Error is detected ---\n');
      }
      break;
    case 'run7':
      console.log('--- run7 is activated ---\n');
      const run7_result = changeWPM(dataList, request.target).then(dataList2 => {writePrev(dataList2);});
      if(run7_result){
        sendResponse(run7_result);
      }
      else{
        sendResponse('--- Error is detected ---\n');
      }
      break;
  }

  return true;
});